/*
 * Functions to implement and manage the keyboard editing operations
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <dirent.h>
#include <unistd.h>

#include "support.h"
#include "interface.h"

#include "global.h"
#include "keyboard.h"

/**********************************************************************
 * Global variables
 */
extern GtkWidget *window_main_;
extern GtkWidget *window_keyboard_;
extern GtkWidget *window_hints_;
extern GtkWidget *window_tutor_;
extern gboolean is_initializing_list;
extern gboolean keyboard_is_modified;
extern gchar *data_path;
extern gchar *kb_name;
extern gchar lowerchars[4][KEY_LINE_LEN + 1];
extern gchar upperchars[4][KEY_LINE_LEN + 1];
extern gchar *consonants;
extern gchar *symbols;

/**********************************************************************
 * Global variables used only here
 */
gchar virtual_key[32];

/**********************************************************************
 * Read the character sets (lowerchars[] & upperchars[])
 * for the keyboard currently selected.
 */
void
get_keyboard ()
{
  gint i;
  gint pos = 0;
  gchar *tmp_name = NULL;
  FILE *fh;
  GtkWidget *wg;

  /*
   * Search at home
   */
  tmp_name =
    g_strconcat (g_get_home_dir (), "/.klavaro/", kb_name, ".kbd", NULL);
  fh = fopen (tmp_name, "r");
  if (fh == NULL)
    {
      /*
       * Search at data
       */
      g_free (tmp_name);
      tmp_name = g_strconcat (data_path, kb_name, ".kbd", NULL);
      fh = fopen (tmp_name, "r");
    }
  g_free (tmp_name);

  if (fh)
    {
      for (i = 0; i < 4; i++)
	{
	  fgets (lowerchars[i], KEY_LINE_LEN + 1, fh);
	}
      for (i = 0; i < 4; i++)
	{
	  fgets (upperchars[i], KEY_LINE_LEN + 1, fh);
	}
      fclose (fh);

      init_consonants_symbols ();
    }
  else
    {
      if (strcmp (kb_name, DEFAULT_KEYBOARD) == 0)
	g_error ("couldn't open the default keyboard layout.");

      wg = lookup_widget (window_main_, "entry_keyboard");
      if (strcmp (kb_name, "") == 0)
	{
	  g_free (kb_name);
	  kb_name = gtk_editable_get_chars (GTK_EDITABLE (wg), 0, -1);
	  get_keyboard ();
	  return;
	}

      g_message ("couldn't find the keyboard layout: \"%s\"\n"
		 " Opening the default one: \"%s\"", kb_name,
		 DEFAULT_KEYBOARD);
      gtk_editable_delete_text (GTK_EDITABLE (wg), 0, -1);
      gtk_editable_insert_text (GTK_EDITABLE (wg),
				DEFAULT_KEYBOARD, strlen (DEFAULT_KEYBOARD),
				&pos);
      g_free (kb_name);
      kb_name = g_strconcat (DEFAULT_KEYBOARD, NULL);
      get_keyboard ();
      return;
    }
}

/**********************************************************************
 * Organize the set of available characters of the keyboard
 */
void
init_consonants_symbols ()
{
  gint i, j;
  gchar chr;
  gchar chr_str[2];

  chr_str[1] = '\0';
  g_free (consonants);
  g_free (symbols);
  consonants = g_strconcat ("", NULL);
  symbols = g_strconcat ("", NULL);
  for (i = 0; i < 4; i++)
    {
      for (j = 0; j < KEY_LINE_LEN; j++)
	{
	  chr_str[0] = chr = lowerchars[i][j];
	  if (isalpha (chr) && chr != 'a' && chr != 'e' && chr != 'i'
	      && chr != 'o' && chr != 'u')
	    consonants = g_strconcat (chr_str, consonants, NULL);
	  else if (ispunct (chr))
	    symbols = g_strconcat (chr_str, symbols, NULL);

	  chr_str[0] = chr = tolower (upperchars[i][j]);
	  if (isalpha (chr) && chr != 'a' && chr != 'e' && chr != 'i'
	      && chr != 'o' && chr != 'u' && chr != lowerchars[i][j])
	    consonants = g_strconcat (chr_str, consonants, NULL);
	  else if (ispunct (chr))
	    symbols = g_strconcat (chr_str, symbols, NULL);
	}
    }
}

/**********************************************************************
 * Save the custom keyboard layout created by the user
 */
void
save_new_layout ()
{
  gint i;
  gchar *tmp_name = NULL;
  FILE *fh;

  assert_user_dir ();
  tmp_name =
    g_strconcat (g_get_home_dir (), "/.klavaro/", kb_name, ".kbd", NULL);
  fh = fopen (tmp_name, "w");
  g_free (tmp_name);

  for (i = 0; i < 4; i++)
    {
      fprintf (fh, "%s", lowerchars[i]);
    }
  for (i = 0; i < 4; i++)
    {
      fprintf (fh, "%s", upperchars[i]);
    }
  fclose (fh);

  gtk_widget_hide (lookup_widget (window_keyboard_, "label_modified"));
  gtk_widget_set_sensitive (lookup_widget (window_keyboard_, "label_kb_load"),
			    TRUE);
  if (strcmp (kb_name, ".tmp") == 0)
    gtk_widget_show (lookup_widget (window_keyboard_, "label_saved_tmp"));
  keyboard_is_modified = FALSE;
}

/**********************************************************************
 * Remove custom keyboard layout created by the user
 */
void
remove_user_layout ()
{
  gchar *aux;
  gchar *tmp_name;
  GtkWidget *wg;

  wg = lookup_widget (window_keyboard_, "entry_kb_edit");
  aux = gtk_editable_get_chars (GTK_EDITABLE (wg), 0, -1);

  tmp_name = g_strconcat (g_get_home_dir (), "/.klavaro/", aux, ".kbd", NULL);
  unlink (tmp_name);
  g_free (tmp_name);

  if (strcmp (kb_name, aux) == 0)
    {
      g_free (kb_name);
      wg = lookup_widget (window_main_, "entry_keyboard");
      kb_name = gtk_editable_get_chars (GTK_EDITABLE (wg), 0, -1);
      get_keyboard ();
      update_virtual_keyboard ();
    }
  fill_keyboard_list (0);

  g_free (aux);
}


/**********************************************************************
 * Update the virtual keyboard accordingly to its character set and
 * shift key state.
 */
void
update_virtual_keyboard ()
{
  gint i, j;
  gchar lab_name[10];
  gchar char_str[2];
  GtkWidget *wg;
  gboolean tog_state;

  strcpy (lab_name, "lab");
  char_str[1] = '\0';
  wg = lookup_widget (window_keyboard_, "toggle_shift1");
  tog_state = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wg));
  for (i = 0; i < 4; i++)
    {
      for (j = 0; j < KEY_LINE_LEN - (i ? 3 : 1); j++)
	{

	  char_str[0] = tog_state ? upperchars[i][j] : lowerchars[i][j];
	  char_str[0] = toupper (char_str[0]);

	  snprintf (lab_name + (i + 3), 3, "%i", j + 1);
	  wg = lookup_widget (window_keyboard_, lab_name);
	  if (wg)
	    {
	      gtk_label_set (GTK_LABEL (wg), char_str);
	    }
	}
      lab_name[i + 3] = '_';
    }
}

/**********************************************************************
 * Reads the lists of available keyboard layouts
 */
void
fill_keyboard_list (gint entry_type)
{
  gboolean kb_name_exists;
  gint i;
  gint name_len;
  gchar *tmp = NULL;
  gchar tmp_end[5];
  gchar tmp_name[MAX_KEYBOARDS][21];
  DIR *dir = NULL;
  struct dirent *dentry = NULL;
  GList *files = NULL;
  GtkWidget *wg = NULL;

  if (entry_type == 1)
    {
      /*
       * Reads the list of original files
       */
      dir = opendir (data_path);
      if (dir == NULL)
	g_error ("could not find the data directory:/n %s/n", data_path);
    }
  else if (entry_type == 2 || entry_type == 3)
    {
      /*
       * Reads the list of custom files
       */
      assert_user_dir ();
      tmp = g_strconcat (g_get_home_dir (), "/.klavaro", NULL);
      dir = opendir (tmp);
      g_free (tmp);
    }
  else if (entry_type == 0)
    {
      /*
       * Reads the list of custom and original files
       */
      fill_keyboard_list (3);
      fill_keyboard_list (2);
      fill_keyboard_list (1);
      return;
    }

  kb_name_exists = FALSE;
  i = 0;
  while (i < MAX_KEYBOARDS && (dentry = readdir (dir)))
    {
      name_len = strlen (dentry->d_name);
      if (name_len > 20)
	continue;
      if (name_len < 5)
	continue;

      strcpy (tmp_end, dentry->d_name + name_len - 4);
      if (strcmp (tmp_end, ".kbd"))
	continue;

      strcpy (tmp_name[i], dentry->d_name);
      tmp_name[i][name_len - 4] = '\0';

      if (entry_type == 3 && strcmp (tmp_name[i], ".tmp") == 0)
	continue;

      if (strcmp (tmp_name[i], kb_name) != 0)
	files = g_list_insert_sorted (files, tmp_name[i++], comp_str);
      else
	kb_name_exists = TRUE;

    }
  closedir (dir);
  if (kb_name_exists)
    files = g_list_insert (files, kb_name, 0);
  else
    files = g_list_insert (files, "", 0);

  is_initializing_list = TRUE;
  if (entry_type == 1)
    {
      wg = lookup_widget (window_keyboard_, "combo_kb_main");
      gtk_combo_set_popdown_strings (GTK_COMBO (wg), files);
    }
  else if (entry_type == 2)
    {
      wg = lookup_widget (window_keyboard_, "combo_kb_user");
      gtk_combo_set_popdown_strings (GTK_COMBO (wg), files);
    }
  else if (entry_type == 3)
    {
      wg = lookup_widget (window_keyboard_, "combo_kb_edit");
      gtk_combo_set_popdown_strings (GTK_COMBO (wg), files);
    }
  is_initializing_list = FALSE;

  g_list_free (files);

  /*
   * Updates the "Layout to be selected" field
   */
  wg = lookup_widget (window_keyboard_, "label_kb_name");
  gtk_label_set (GTK_LABEL (wg), kb_name);

}

/**********************************************************************
 * Create a popup window to get a real key pressed and
 * write the name of the virtual one.
 */
void
edit_key (GtkButton * button, const gchar * pix_file_name)
{
  gchar pix_name_aux[32];
  GtkWidget *popup_keyedit;
  GtkWidget *pix;
  GtkBox *box;

  strcpy (virtual_key, gtk_widget_get_name (GTK_WIDGET (button)));

  /*
   * Tutoring mode
   */
  if (window_tutor_ != NULL)
    {
      pix = lookup_widget (window_keyboard_, "pix_hands");
      box = GTK_BOX (lookup_widget (window_keyboard_, "hbox_kb_tutor"));
      gtk_widget_destroy (pix);

      strcpy (pix_name_aux, pix_file_name);
      get_hints_from_label (pix_name_aux, virtual_key);
      pix = create_pixmap (window_keyboard_, pix_name_aux);
      gtk_widget_set_name (pix, "pix_hands");
      gtk_widget_ref (pix);
      gtk_object_set_data_full (GTK_OBJECT (window_keyboard_), "pix_hands",
				pix, (GtkDestroyNotify) gtk_widget_unref);
      gtk_box_pack_start (box, pix, TRUE, TRUE, 0);
      gtk_misc_set_alignment (GTK_MISC (pix), 0, 0.5);
      gtk_box_reorder_child (box, pix, 1);
      gtk_widget_show (pix);
      return;
    }

  /*
   * Editing mode
   */
  virtual_key[0] = 'l';
  virtual_key[1] = 'a';
  virtual_key[2] = 'b';
  popup_keyedit = create_popup_keyedit ();
  gtk_widget_show (popup_keyedit);
}

/**********************************************************************
 * Apply the key pressed to the virtual keyboard
 * and to the upper or lower character sets
 */
void
change_key (gchar * real_key)
{
  gint key_lin, key_col;
  gchar str_char[2];
  gboolean tog_state;
  GtkWidget *wg;

  str_char[0] = toupper (real_key[0]);
  str_char[1] = '\0';
  wg = lookup_widget (window_keyboard_, virtual_key);
  gtk_label_set (GTK_LABEL (wg), str_char);

  for (key_lin = 0; virtual_key[key_lin + 3] == '_'; key_lin++);
  key_col = atoi (virtual_key + (key_lin + 3)) - 1;

  wg = lookup_widget (window_keyboard_, "toggle_shift1");
  tog_state = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wg));
  if (tog_state)
    {
      upperchars[key_lin][key_col] = str_char[0];
      if (str_char[0] >= 'A' && str_char[0] <= 'Z')
	lowerchars[key_lin][key_col] = tolower (str_char[0]);
    }
  else
    {
      lowerchars[key_lin][key_col] = tolower (str_char[0]);
      if (str_char[0] >= 'A' && str_char[0] <= 'Z')
	upperchars[key_lin][key_col] = str_char[0];
    }

  if (keyboard_is_modified == FALSE)
    {
      keyboard_is_modified = TRUE;
      gtk_widget_hide (lookup_widget (window_keyboard_, "label_saved_tmp"));
      gtk_widget_show (lookup_widget (window_keyboard_, "label_modified"));
      gtk_widget_set_sensitive (lookup_widget (window_keyboard_,
					       "label_kb_load"), FALSE);
      wg = lookup_widget (window_keyboard_, "notebook_kb");
      gtk_notebook_set_page (GTK_NOTEBOOK (wg), 1);
    }
}

/*******************************************************************************
 * Initialize the hints mapping array
 */

gchar hints[4][KEY_LINE_LEN + 1];
gboolean hints_is_initialized = FALSE;

void
hints_init ()
{
  gint i;
  gchar *tmp_name;
  FILE *fh;

  if (hints_is_initialized == TRUE)
    return;

  hints_is_initialized = TRUE;
  tmp_name = g_strconcat (data_path, "fingers_position.txt", NULL);
  fh = fopen (tmp_name, "r");
  if (fh)
    {
      for (i = 0; i < 4; i++)
	fgets (hints[i], KEY_LINE_LEN + 1, fh);
      fclose (fh);
    }
  else
    {
      g_warning ("couldn't open the file:\n %s", tmp_name);
      g_free (tmp_name);
      return;
    }
  g_free (tmp_name);
}

/**********************************************************************
 * Map the character to the file which shows the finger associated with that key
 */
void
get_hands_hint_from_char (gchar * file_name, gchar character)
{
  gint i;
  gint j;

  strcpy (file_name, "hands_9.xpm");
  if (character == '\n' || character == (gchar) 182)
    return;
  strcpy (file_name, "hands_5.xpm");
  if (character == ' ')
    return;

  hints_init ();

  for (i = 3; i >= 0; i--)
    for (j = 0; j < 15; j++)
      if (character == lowerchars[i][j])
	{
	  file_name[6] = hints[i][j];
	  return;
	}

  for (i = 3; i >= 0; i--)
    for (j = 0; j < 15; j++)
      if (character == upperchars[i][j])
	{
	  file_name[6] = hints[i][j];
	  return;
	}
  file_name[6] = '0';
}

/**********************************************************************
 * Map the label to the file which shows the finger associated with that key
 */
void
get_hints_from_label (gchar * file_name, gchar * label)
{
  gint key_lin;
  gint key_col;
  gchar *tailptr;

  if (strstr (label, "but_space") != NULL)
    return;
  if (strstr (label, "but") == NULL)
    return;

  strcpy (file_name, "hands_0.xpm");

  for (key_lin = 0; label[key_lin + 3] == '_'; key_lin++);
  key_col = strtol (label + (key_lin + 3), &tailptr, 0) - 1;

  if (*tailptr != '\0')
    return;

  hints_init ();

  file_name[6] = hints[key_lin][key_col];
}

/**********************************************************************
 * Update the image of the window_hints_
 */
void
update_hints (gchar character)
{
  gchar file_name[32];
  GtkWidget *wg;
  GtkWidget *pix;

  if (window_hints_ == NULL)
    return;

  wg = lookup_widget (window_hints_, "pixmap_hints");
  gtk_widget_destroy (wg);

  get_hands_hint_from_char (file_name, character);
  pix = create_pixmap (window_hints_, file_name);
  gtk_widget_set_name (pix, "pixmap_hints");
  gtk_widget_ref (pix);
  gtk_object_set_data_full (GTK_OBJECT (window_hints_), "pixmap_hints", pix,
			    (GtkDestroyNotify) gtk_widget_unref);
  wg = lookup_widget (window_hints_, "frame_hints");
  gtk_container_add (GTK_CONTAINER (wg), pix);
  gtk_widget_show (pix);
}

/*
 * A hack to switch the note tab
 */
gboolean
keyb_force_edit_tab (gpointer data)
{
  GtkWidget *wg = NULL;

  gdk_beep ();
  wg = lookup_widget (window_keyboard_, "notebook_kb");

  if (wg == NULL)
    return (FALSE);

  gtk_notebook_set_page (GTK_NOTEBOOK (wg), 1);
  return (FALSE);
}
